@@ -1,6 +1,6 @@
 from __future__ import print_function
 from numpy import asarray, unravel_index, prod, mod, ndarray, ceil, where, \
-    r_, sort, argsort, array, random, arange, ones, expand_dims
+    r_, sort, argsort, array, random, arange, ones, expand_dims, sum
 from itertools import groupby
 
 from bolt.base import BoltArray
@@ -12,14 +12,20 @@ from bolt.utils import slicify, listify, tupleize, argpack, inshape, istranspose
 
 class BoltArraySpark(BoltArray):
 
-    _metadata = BoltArray._metadata + ['_shape', '_split', '_dtype']
+    _metadata = {
+        '_shape': None,
+        '_split': None,
+        '_dtype': None,
+        '_ordered': True
+    }
 
-    def __init__(self, rdd, shape=None, split=None, dtype=None):
+    def __init__(self, rdd, shape=None, split=None, dtype=None, ordered=True):
         self._rdd = rdd
         self._shape = shape
         self._split = split
         self._dtype = dtype
         self._mode = 'spark'
+        self._ordered = ordered
 
     @property
     def _constructor(self):
@@ -102,7 +108,7 @@ class BoltArraySpark(BoltArray):
         from bolt.local.array import BoltArrayLocal
         return BoltArrayLocal(self._rdd.values().first())
 
-    def map(self, func, axis=(0,), value_shape=None):
+    def map(self, func, axis=(0,), value_shape=None, with_keys=False):
         """
         Apply a function across an axis.
 
@@ -112,7 +118,8 @@ class BoltArraySpark(BoltArray):
         Parameters
         ----------
         func : function
-            Function of a single array to apply
+            Function of a single array to apply. If with_keys=True,
+            function should be of a (tuple, array) pair.
 
         axis : tuple or int, optional, default=(0,)
             Axis or multiple axes to apply function along.
@@ -120,6 +127,9 @@ class BoltArraySpark(BoltArray):
         value_shape : tuple, optional, default=None
             Known shape of values resulting from operation
 
+        with_keys : bool, optional, default=False
+            Include keys as an argument to the function
+
         Returns
         -------
         BoltArraySpark
@@ -127,21 +137,29 @@ class BoltArraySpark(BoltArray):
         axis = tupleize(axis)
         swapped = self._align(axis)
 
+        if with_keys:
+            test_func = lambda x: func(((0,), x))
+        else:
+            test_func = func
+
         if value_shape is None:
             # try to compute the size of each mapped element by applying func to a random array
             value_shape = None
             try:
-                value_shape = func(random.randn(*swapped.values.shape).astype(self.dtype)).shape
+                value_shape = test_func(random.randn(*swapped.values.shape).astype(self.dtype)).shape
             except Exception:
                 first = swapped._rdd.first()
                 if first:
                     # eval func on the first element
-                    mapped = func(first[1])
+                    mapped = test_func(first[1])
                     value_shape = mapped.shape
 
         shape = tuple([swapped._shape[ax] for ax in range(len(axis))]) + tupleize(value_shape)
 
-        rdd = swapped._rdd.mapValues(func)
+        if with_keys:
+            rdd = swapped._rdd.map(lambda kv: (kv[0], func(kv)))
+        else:
+            rdd = swapped._rdd.mapValues(func)
 
         # reshaping will fail if the elements aren't uniformly shaped
         def check(v):
@@ -186,7 +204,7 @@ class BoltArraySpark(BoltArray):
         count, zipped = zip_with_index(rdd)
         if not count:
             count = zipped.count()
-        reindexed = zipped.map(lambda kv: (kv[1], kv[0]))
+        reindexed = zipped.map(lambda kv: (tupleize(kv[1]), kv[0]))
 
         # since we can only filter over one axis, the remaining shape is always the following
         remaining = list(swapped.shape[1:])
@@ -223,7 +241,7 @@ class BoltArraySpark(BoltArray):
 
         axis = tupleize(axis)
         swapped = self._align(axis)
-        arr = swapped._rdd.values().reduce(func)
+        arr = swapped._rdd.values().treeReduce(func, depth=3)
 
         if keepdims:
             for i in axis:
@@ -277,7 +295,7 @@ class BoltArraySpark(BoltArray):
 
             counter = swapped._rdd.values()\
                              .mapPartitions(lambda i: [StatCounter(values=i, stats=name)])\
-                             .reduce(reducer)
+                             .treeReduce(reducer, depth=3)
 
             arr = getattr(counter, name)
 
@@ -432,7 +450,7 @@ class BoltArraySpark(BoltArray):
         shape = tuple([x + y if i == axis else x
                       for i, (x, y) in enumerate(zip(self.shape, arry.shape))])
 
-        return self._constructor(rdd, shape=shape).__finalize__(self)
+        return self._constructor(rdd, shape=shape, ordered=False).__finalize__(self)
 
     def _getbasic(self, index):
         """
@@ -500,6 +518,43 @@ class BoltArraySpark(BoltArray):
 
         return rdd, shape, split
 
+    def _getmixed(self, index):
+        """
+        Mixed indexing (combines basic and advanced indexes)
+
+        Assumes that only a single advanced index is used, due to the complicated
+        behavior needed to be compatible with NumPy otherwise.
+        """
+        # find the single advanced index
+        index = list(index)
+        loc = where([isinstance(i, (tuple, list, ndarray)) for i in index])[0][0]
+        idx = list(index[loc])
+
+        if isinstance(idx[0], (tuple, list, ndarray)):
+            raise ValueError("When mixing basic and advanced indexing, "
+                             "advanced index must be one-dimensional")
+
+        # single advanced index is on a key -- filter and update key
+        if loc < self.split:
+            def newkey(key):
+                newkey = list(key)
+                newkey[loc] = idx.index(key[loc])
+                return tuple(newkey)
+            rdd = self._rdd.filter(lambda kv: kv[0][loc] in idx).map(lambda kv: (newkey(kv[0]), kv[1]))
+        # single advanced index is on a value -- use NumPy indexing
+        else:
+            slices = [slice(0, size) for size in self.shape[self.split:]]
+            slices[loc - self.split] = idx
+            rdd = self._rdd.map(lambda kv: (kv[0], kv[1][slices]))
+        newshape = list(self.shape)
+        newshape[loc] = len(idx)
+        barray = self._constructor(rdd, shape=tuple(newshape)).__finalize__(self)
+
+        # apply the rest of the simple indices
+        index[loc] = slice(None)
+        barray = barray[index]
+        return barray._rdd, barray.shape, barray.split
+
     def __getitem__(self, index):
         """
         Get an item from the array through indexing.
@@ -511,25 +566,51 @@ class BoltArraySpark(BoltArray):
 
         Parameters
         ----------
-        index : tuple of slices, ints, list, sets, or ndarrays
+        index : tuple of slices, ints, list, tuple, or ndarrays
             One or more index specifications
 
         Returns
         -------
         BoltSparkArray
         """
-        index = tupleize(index)
+        index = list(tupleize(index))
 
         if len(index) > self.ndim:
             raise ValueError("Too many indices for array")
 
-        if not all([isinstance(i, (slice, int, list, set, ndarray)) for i in index]):
+        if not all([isinstance(i, (slice, int, list, tuple, ndarray)) for i in index]):
             raise ValueError("Each index must either be a slice, int, list, set, or ndarray")
 
         # fill unspecified axes with full slices
         if len(index) < self.ndim:
             index += tuple([slice(0, None, None) for _ in range(self.ndim - len(index))])
 
+        # check bounds
+        for n, idx in enumerate(index):
+            size = self.shape[n]
+            if isinstance(idx, slice):
+                minval = 0 if idx.start is None else idx.start
+                maxval = size if idx.stop is None else idx.stop
+                # account for negative indices
+                if minval < 0: minval = size + minval
+                if maxval < 0: maxval = size + maxval
+                # account for over-flowing the bounds
+                if minval < 0: minval = 0
+                if maxval > size: maxval = size
+                # throw an error if this would lead to an empty dimension in numpy
+                if minval > size-1 or maxval < 1 or minval >= maxval:
+                    raise ValueError("Index {} in in dimension {} with shape {} would "
+                                     "produce an empty dimension".format(idx, n, size))
+                else:
+                    index[n] = slice(minval, maxval, idx.step)
+            else:
+                pos = [size + i if i<0 else i for i in array(idx).flatten()]
+                minval = min(pos)
+                maxval = max(pos)
+                if minval < 0 or maxval > size-1:
+                    raise ValueError("Index {} out of bounds in dimension {} with "
+                                     "shape {}".format(idx, n, size))
+
         # convert ints to lists if not all ints and slices
         if not all([isinstance(i, (int, slice)) for i in index]):
             index = tuple([[i] if isinstance(i, int) else i for i in index])
@@ -537,11 +618,14 @@ class BoltArraySpark(BoltArray):
         # select basic or advanced indexing
         if all([isinstance(i, (slice, int)) for i in index]):
             rdd, shape, split = self._getbasic(index)
-        elif all([isinstance(i, (set, list, ndarray)) for i in index]):
+        elif all([isinstance(i, (tuple, list, ndarray)) for i in index]):
             rdd, shape, split = self._getadvanced(index)
+        elif sum([isinstance(i, (tuple, list, ndarray)) for i in index]) == 1:
+            rdd, shape, split = self._getmixed(index)
         else:
-            raise NotImplementedError("Cannot mix basic indexing (slices and ints) with "
-                                      "advanced indexing (lists and ndarrays) across axes")
+            raise NotImplementedError("When mixing basic indexing (slices and int) with "
+                                      "with advanced indexing (lists, tuples, and ndarrays), "
+                                      "can only have a single advanced index")
 
         result = self._constructor(rdd, shape=shape, split=split).__finalize__(self)
 
@@ -552,7 +636,7 @@ class BoltArraySpark(BoltArray):
             tosqueeze = tuple([k for k, i in enumerate(index) if isinstance(i, int)])
             return result.squeeze(tosqueeze)
 
-    def chunk(self, size="150", axis=None):
+    def chunk(self, size="150", axis=None, padding=None):
         """
         Chunks records of a distributed array.
 
@@ -571,6 +655,11 @@ class BoltArraySpark(BoltArray):
             One or more axis to chunk array along, if None
             will use all axes,
 
+        padding: tuple or int, default = None
+            Number of elements per dimension that will overlap with the adjacent chunk.
+            If a tuple, specifies padding along each chunked dimension; if a int, same
+            padding will be applied to all chunked dimensions.
+
         Returns
         -------
         ChunkedArray
@@ -578,11 +667,12 @@ class BoltArraySpark(BoltArray):
         if type(size) is not str:
             size = tupleize((size))
         axis = tupleize((axis))
+        padding = tupleize((padding))
 
         from bolt.spark.chunk import ChunkedArray
 
         chnk = ChunkedArray(rdd=self._rdd, shape=self._shape, split=self._split, dtype=self._dtype)
-        return chnk._chunk(size, axis)
+        return chnk._chunk(size, axis, padding)
 
     def swap(self, kaxes, vaxes, size="150"):
         """
@@ -592,7 +682,7 @@ class BoltArraySpark(BoltArray):
         on the Spark bolt array. It exchanges an arbitrary set of axes
         between the keys and the valeus. If either is None, will only
         move axes in one direction (from keys to values, or values to keys).
-        Keys moved to values will be placed immediately after the split; 
+        Keys moved to values will be placed immediately after the split;
         values moved to keys will be placed immediately before the split.
 
         Parameters
@@ -662,7 +752,7 @@ class BoltArraySpark(BoltArray):
         swapping_values = sort(new_keys[new_keys >= split])
         stationary_keys = sort(new_keys[new_keys < split])
         stationary_values = sort(new_values[new_values >= split])
-        
+
         # compute the permutation that the swap causes
         p_swap = r_[stationary_keys, swapping_values, swapping_keys, stationary_values]
 
@@ -676,7 +766,7 @@ class BoltArraySpark(BoltArray):
         arr = self.swap(swapping_keys, swapping_values-split)
         arr = arr.keys.transpose(tuple(p_keys.tolist()))
         arr = arr.values.transpose(tuple(p_values.tolist()))
-        
+
         return arr
 
     @property
@@ -881,7 +971,8 @@ class BoltArraySpark(BoltArray):
 
         Will likely cause memory problems for large objects.
         """
-        x = self._rdd.sortByKey().values().collect()
+        rdd = self._rdd if self._ordered else self._rdd.sortByKey()
+        x = rdd.values().collect()
         return asarray(x).reshape(self.shape)
 
     def tordd(self):
@@ -896,5 +987,3 @@ class BoltArraySpark(BoltArray):
         """
         for x in self._rdd.take(10):
             print(x)
-
-
